@@ -18,8 +18,12 @@ typedef struct DispatcherObject{
     char can_compile;        /* Can auto compile */
     /* Borrowed references */
     PyObject *firstdef, *fallbackdef, *interpdef;
+    /* Whether to fold named arguments and default values (false for lifted loops)*/
+    int fold_args;
     /* Tuple of argument names */
     PyObject *argnames;
+    /* Tuple of default values */
+    PyObject *defargs;
 } DispatcherObject;
 
 static int tc_int8;
@@ -93,28 +97,38 @@ PyObject* init_types(PyObject *self, PyObject *args)
     Py_RETURN_NONE;
 }
 
+static int
+Dispatcher_traverse(DispatcherObject *self, visitproc visit, void *arg)
+{
+    Py_VISIT(self->defargs);
+    return 0;
+}
+
 static void
 Dispatcher_dealloc(DispatcherObject *self)
 {
-    Py_DECREF(self->argnames);
+    Py_XDECREF(self->argnames);
+    Py_XDECREF(self->defargs);
     dispatcher_del(self->dispatcher);
     Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
 
-static
-int
+static int
 Dispatcher_init(DispatcherObject *self, PyObject *args, PyObject *kwds)
 {
     PyObject *tmaddrobj;
     void *tmaddr;
     int argct;
 
-    if (!PyArg_ParseTuple(args, "OiO!",
-                          &tmaddrobj, &argct, &PyTuple_Type, &self->argnames)) {
+    if (!PyArg_ParseTuple(args, "OiiO!O!", &tmaddrobj, &argct,
+                          &self->fold_args,
+                          &PyTuple_Type, &self->argnames,
+                          &PyTuple_Type, &self->defargs)) {
         return -1;
     }
     Py_INCREF(self->argnames);
+    Py_INCREF(self->defargs);
     tmaddr = PyLong_AsVoidPtr(tmaddrobj);
     self->dispatcher = dispatcher_new(tmaddr, argct);
     self->can_compile = 1;
@@ -442,20 +456,36 @@ int typecode(DispatcherObject *dispatcher, PyObject *val) {
 }
 
 static
-void explain_ambiguous(PyObject *dispatcher, PyObject *args, PyObject *kws) {
+void explain_issue(PyObject *dispatcher, PyObject *args, PyObject *kws,
+                   const char *method_name, const char *default_msg)
+{
     PyObject *callback, *result;
-    callback = PyObject_GetAttrString(dispatcher, "_explain_ambiguous");
+    callback = PyObject_GetAttrString(dispatcher, method_name);
     if (!callback) {
-        PyErr_SetString(PyExc_TypeError, "Ambigous overloading");
+        PyErr_SetString(PyExc_TypeError, default_msg);
         return;
     }
     result = PyObject_Call(callback, args, kws);
+    Py_DECREF(callback);
     if (result != NULL) {
-        PyErr_SetString(PyExc_RuntimeError,
-                        "_explain_ambiguous must raise an exception");
+        PyErr_Format(PyExc_RuntimeError, "%s must raise an exception",
+                     method_name);
         Py_DECREF(result);
     }
-    Py_XDECREF(callback);
+}
+
+static
+void explain_ambiguous(PyObject *dispatcher, PyObject *args, PyObject *kws)
+{
+    explain_issue(dispatcher, args, kws, "_explain_ambiguous",
+                  "Ambigous overloading");
+}
+
+static
+void explain_matching_error(PyObject *dispatcher, PyObject *args, PyObject *kws)
+{
+    explain_issue(dispatcher, args, kws, "_explain_matching_error",
+                  "No matching definition");
 }
 
 /* A custom, fast, inlinable version of PyCFunction_Call() */
@@ -506,19 +536,37 @@ find_named_args(DispatcherObject *self, PyObject **pargs, PyObject **pkws)
     PyObject *kws = *pkws;
     Py_ssize_t pos_args = PyTuple_GET_SIZE(oldargs);
     Py_ssize_t named_args, total_args, i;
+    Py_ssize_t func_args = PyTuple_GET_SIZE(self->argnames);
+    Py_ssize_t defaults = PyTuple_GET_SIZE(self->defargs);
+    Py_ssize_t defstart = func_args - defaults;
 
-    if (kws == NULL || (named_args = PyDict_Size(kws)) == 0) {
+    if (kws != NULL)
+        named_args = PyDict_Size(kws);
+    else
+        named_args = 0;
+    if (named_args == 0 && pos_args == func_args) {
         Py_INCREF(oldargs);
         return 0;
     }
     total_args = pos_args + named_args;
-    if (total_args > PyTuple_GET_SIZE(self->argnames)) {
+    if (total_args > func_args) {
         PyErr_Format(PyExc_TypeError,
                      "too many arguments: expected %d, got %d",
-                     (int) PyTuple_GET_SIZE(self->argnames), (int) total_args);
+                     (int) func_args, (int) total_args);
         return -1;
     }
-    newargs = PyTuple_New(total_args);
+    else if (total_args < func_args - defaults) {
+        if (defaults == 0)
+            PyErr_Format(PyExc_TypeError,
+                         "not enough arguments: expected %d, got %d",
+                         (int) func_args, (int) total_args);
+        else
+            PyErr_Format(PyExc_TypeError,
+                         "not enough arguments: expected at least %d, got %d",
+                         (int) (func_args - defaults), (int) total_args);
+        return -1;
+    }
+    newargs = PyTuple_New(func_args);
     if (!newargs)
         return -1;
     for (i = 0; i < pos_args; i++) {
@@ -526,18 +574,31 @@ find_named_args(DispatcherObject *self, PyObject **pargs, PyObject **pkws)
         Py_INCREF(value);
         PyTuple_SET_ITEM(newargs, i, value);
     }
-    for (i = pos_args; i < total_args; i++) {
+    /* Iterate over missing positional arguments, try to find them in
+       named arguments or default values. */
+    for (i = pos_args; i < func_args; i++) {
         PyObject *name = PyTuple_GET_ITEM(self->argnames, i);
-        PyObject *value = PyDict_GetItem(kws, name);
-        if (value == NULL) {
-            PyErr_Format(PyExc_TypeError,
-                         "missing argument '%s'",
-                         PyString_AsString(name));
-            Py_DECREF(newargs);
-            return -1;
+        if (kws != NULL) {
+            /* Named argument? */
+            PyObject *value = PyDict_GetItem(kws, name);
+            if (value != NULL) {
+                Py_INCREF(value);
+                PyTuple_SET_ITEM(newargs, i, value);
+                continue;
+            }
         }
-        Py_INCREF(value);
-        PyTuple_SET_ITEM(newargs, i, value);
+        if (i >= defstart) {
+            /* Argument has a default value? */
+            PyObject *value = PyTuple_GET_ITEM(self->defargs, i - defstart);
+            Py_INCREF(value);
+            PyTuple_SET_ITEM(newargs, i, value);
+            continue;
+        }
+        PyErr_Format(PyExc_TypeError,
+                     "missing argument '%s'",
+                     PyString_AsString(name));
+        Py_DECREF(newargs);
+        return -1;
     }
     *pargs = newargs;
     *pkws = NULL;
@@ -555,8 +616,12 @@ Dispatcher_call(DispatcherObject *self, PyObject *args, PyObject *kws)
     int matches;
     PyObject *cfunc;
 
-    if (find_named_args(self, &args, &kws))
-        return NULL;
+    if (self->fold_args) {
+        if (find_named_args(self, &args, &kws))
+            return NULL;
+    }
+    else
+        Py_INCREF(args);
     /* Now we own a reference to args */
 
     argct = PySequence_Fast_GET_SIZE(args);
@@ -590,7 +655,7 @@ Dispatcher_call(DispatcherObject *self, PyObject *args, PyObject *kws)
             retval = call_cfunc(self->fallbackdef, args, kws);
         } else {
             /* Raise TypeError */
-            PyErr_SetString(PyExc_TypeError, "No matching definition");
+            explain_matching_error((PyObject *) self, args, kws);
             retval = NULL;
         }
     } else if (self->can_compile) {
@@ -598,7 +663,7 @@ Dispatcher_call(DispatcherObject *self, PyObject *args, PyObject *kws)
         retval = compile_and_invoke(self, args, kws);
     } else {
         /* Ambiguous */
-        explain_ambiguous((PyObject*)self, args, kws);
+        explain_ambiguous((PyObject *) self, args, kws);
         retval = NULL;
     }
 
@@ -626,47 +691,47 @@ static PyMemberDef Dispatcher_members[] = {
 static PyTypeObject DispatcherType = {
 #if (PY_MAJOR_VERSION < 3)
     PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    0,                                           /* ob_size */
 #else
     PyVarObject_HEAD_INIT(NULL, 0)
 #endif
-    "_dispatcher.Dispatcher",        /*tp_name*/
-    sizeof(DispatcherObject),     /*tp_basicsize*/
-    0,                         /*tp_itemsize*/
-    (destructor)Dispatcher_dealloc,                         /*tp_dealloc*/
-    0,                         /*tp_print*/
-    0,                         /*tp_getattr*/
-    0,                         /*tp_setattr*/
-    0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
-    0,                         /*tp_as_number*/
-    0,                         /*tp_as_sequence*/
-    0,                         /*tp_as_mapping*/
-    0,                         /*tp_hash */
-    (PyCFunctionWithKeywords)Dispatcher_call, /*tp_call*/
-    0,                         /*tp_str*/
-    0,                         /*tp_getattro*/
-    0,                         /*tp_setattro*/
-    0,                         /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
-    "Dispatcher object",       /* tp_doc */
-    0,                         /* tp_traverse */
-    0,                         /* tp_clear */
-    0,                         /* tp_richcompare */
-    0,                         /* tp_weaklistoffset */
-    0,                         /* tp_iter */
-    0,                         /* tp_iternext */
-    Dispatcher_methods,        /* tp_methods */
-    Dispatcher_members,        /* tp_members */
-    0,                         /* tp_getset */
-    0,                         /* tp_base */
-    0,                         /* tp_dict */
-    0,                         /* tp_descr_get */
-    0,                         /* tp_descr_set */
-    0,                         /* tp_dictoffset */
-    (initproc)Dispatcher_init, /* tp_init */
-    0,                         /* tp_alloc */
-    0,                         /* tp_new */
+    "_dispatcher.Dispatcher",                    /* tp_name */
+    sizeof(DispatcherObject),                    /* tp_basicsize */
+    0,                                           /* tp_itemsize */
+    (destructor)Dispatcher_dealloc,              /* tp_dealloc */
+    0,                                           /* tp_print */
+    0,                                           /* tp_getattr */
+    0,                                           /* tp_setattr */
+    0,                                           /* tp_compare */
+    0,                                           /* tp_repr */
+    0,                                           /* tp_as_number */
+    0,                                           /* tp_as_sequence */
+    0,                                           /* tp_as_mapping */
+    0,                                           /* tp_hash */
+    (PyCFunctionWithKeywords)Dispatcher_call,    /* tp_call*/
+    0,                                           /* tp_str*/
+    0,                                           /* tp_getattro*/
+    0,                                           /* tp_setattro*/
+    0,                                           /* tp_as_buffer*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /* tp_flags*/
+    "Dispatcher object",                         /* tp_doc */
+    (traverseproc) Dispatcher_traverse,          /* tp_traverse */
+    0,                                           /* tp_clear */
+    0,                                           /* tp_richcompare */
+    0,                                           /* tp_weaklistoffset */
+    0,                                           /* tp_iter */
+    0,                                           /* tp_iternext */
+    Dispatcher_methods,                          /* tp_methods */
+    Dispatcher_members,                          /* tp_members */
+    0,                                           /* tp_getset */
+    0,                                           /* tp_base */
+    0,                                           /* tp_dict */
+    0,                                           /* tp_descr_get */
+    0,                                           /* tp_descr_set */
+    0,                                           /* tp_dictoffset */
+    (initproc)Dispatcher_init,                   /* tp_init */
+    0,                                           /* tp_alloc */
+    0,                                           /* tp_new */
 };
 
 
