@@ -1,14 +1,17 @@
 from __future__ import print_function, division, absolute_import
-from pprint import pprint
+
+import inspect
 from contextlib import contextmanager
 from collections import namedtuple, defaultdict
+from pprint import pprint
+import sys
 import warnings
-import inspect
 
-from numba import (bytecode, interpreter, typing, typeinfer, lowering,
-                   objmode, irpasses, utils, config, type_annotations,
+from numba import (bytecode, interpreter, funcdesc, typing, typeinfer,
+                   lowering, objmode, irpasses, utils, config,
                    types, ir, assume, looplifting, macro, types)
 from numba.targets import cpu
+from numba.annotations import type_annotations
 
 
 class Flags(utils.ConfigOptions):
@@ -47,7 +50,7 @@ CR_FIELDS = ["typing_context",
              "fndesc",
              "interpmode",
              "library",
-             "exception_map",
+             "call_helper",
              "environment"]
 
 
@@ -59,7 +62,7 @@ DEFAULT_FUNCTION_ATTRIBUTES = FunctionAttributes('<anonymous>', '<unknown>', 0)
 
 _LowerResult = namedtuple("_LowerResult", [
     "fndesc",
-    "exception_map",
+    "call_helper",
     "cfunc",
     "env",
 ])
@@ -106,6 +109,11 @@ def compile_result(**kws):
     missing = fieldset - keys
     for k in missing:
         kws[k] = None
+    # Avoid keeping alive traceback variables
+    if sys.version_info >= (3,):
+        err = kws['typing_error']
+        if err is not None:
+            kws['typing_error'] = err.with_traceback(None)
     return CompileResult(**kws)
 
 
@@ -292,11 +300,12 @@ class Pipeline(object):
 
         return self.compile_bytecode(bc, func_attr=self.func_attr)
 
-    def compile_bytecode(self, bc, lifted=(),
+    def compile_bytecode(self, bc, lifted=(), lifted_from=None,
                          func_attr=DEFAULT_FUNCTION_ATTRIBUTES):
         self.bc = bc
         self.func = bc.func
         self.lifted = lifted
+        self.lifted_from = lifted_from
         self.func_attr = func_attr
         return self._compile_bytecode()
 
@@ -354,7 +363,7 @@ class Pipeline(object):
             cres = compile_bytecode(self.typingctx, self.targetctx, entry,
                                     self.args, self.return_type,
                                     outer_flags, self.locals,
-                                    lifted=tuple(loops),
+                                    lifted=tuple(loops), lifted_from=None,
                                     func_attr=self.func_attr)
             return cres
 
@@ -401,12 +410,19 @@ class Pipeline(object):
             interp=self.interp,
             typemap=self.typemap,
             calltypes=self.calltypes,
-            lifted=self.lifted)
+            lifted=self.lifted,
+            lifted_from=self.lifted_from,
+            args=self.args,
+            return_type=self.return_type,
+            func_attr=self.func_attr,
+            html_output=config.HTML)
 
         if config.ANNOTATE:
             print("ANNOTATION".center(80, '-'))
             print(self.type_annotation)
             print('=' * 80)
+        if config.HTML:
+            self.type_annotation.html_annotate()
 
     def backend_object_mode(self):
         """
@@ -452,7 +468,7 @@ class Pipeline(object):
                             typing_error=self.status.fail_reason,
                             type_annotation=self.type_annotation,
                             library=self.library,
-                            exception_map=lowered.exception_map,
+                            call_helper=lowered.call_helper,
                             signature=signature,
                             objectmode=objectmode,
                             interpmode=False,
@@ -548,12 +564,12 @@ def compile_extra(typingctx, targetctx, func, args, return_type, flags,
 
 
 def compile_bytecode(typingctx, targetctx, bc, args, return_type, flags,
-                     locals, lifted=(),
+                     locals, lifted=(), lifted_from=None,
                      func_attr=DEFAULT_FUNCTION_ATTRIBUTES, library=None):
 
     pipeline = Pipeline(typingctx, targetctx, library,
                         args, return_type, flags, locals)
-    return pipeline.compile_bytecode(bc=bc, lifted=lifted, func_attr=func_attr)
+    return pipeline.compile_bytecode(bc=bc, lifted=lifted, lifted_from=lifted_from, func_attr=func_attr)
 
 
 def compile_internal(typingctx, targetctx, library,
@@ -589,17 +605,20 @@ def legalize_return_type(return_type, interp, targetctx):
     assert assume.return_argument_array_only
 
     if isinstance(return_type, types.Array):
-        # Walk IR to discover all return statements
+        # Walk IR to discover all arguments and all return statements
         retstmts = []
         caststmts = {}
+        argvars = set()
         for bid, blk in interp.blocks.items():
             for inst in blk.body:
                 if isinstance(inst, ir.Return):
                     retstmts.append(inst.value.name)
-                if (isinstance(inst, ir.Assign)
-                        and isinstance(inst.value, ir.Expr)
+                elif isinstance(inst, ir.Assign):
+                    if (isinstance(inst.value, ir.Expr)
                         and inst.value.op == 'cast'):
-                    caststmts[inst.target.name] = inst.value
+                        caststmts[inst.target.name] = inst.value
+                    elif isinstance(inst.value, ir.Arg):
+                        argvars.add(inst.target.name)
 
         assert retstmts, "No return statements?"
 
@@ -608,11 +627,10 @@ def legalize_return_type(return_type, interp, targetctx):
         #        must be statically resolvable.
 
         # The return value must be the first modification of the value.
-        arguments = set("{0}.1".format(a) for a in interp.argspec.args)
 
         for var in retstmts:
             cast = caststmts.get(var)
-            if cast is None or cast.value.name not in arguments:
+            if cast is None or cast.value.name not in argvars:
                 raise TypeError("Only accept returning of array passed into the "
                                 "function as argument")
 
@@ -632,6 +650,9 @@ def translate_stage(bytecode):
     if config.DEBUG or config.DUMP_IR:
         print(("IR DUMP: %s" % interp.bytecode.func_qualname).center(80, "-"))
         interp.dump()
+        if interp.generator_info:
+            print(("GENERATOR INFO: %s" % interp.bytecode.func_qualname).center(80, "-"))
+            interp.dump_generator_info()
 
     expanded = macro.expand_macros(interp.blocks)
 
@@ -647,11 +668,11 @@ def type_inference_stage(typingctx, interp, args, return_type, locals={}):
     if len(args) != len(interp.argspec.args):
         raise TypeError("Mismatch number of argument types")
 
-    infer = typeinfer.TypeInferer(typingctx, interp.blocks)
+    infer = typeinfer.TypeInferer(typingctx, interp)
 
     # Seed argument types
-    for arg, ty in zip(interp.argspec.args, args):
-        infer.seed_type(arg, ty)
+    for index, (name, ty) in enumerate(zip(interp.argspec.args, args)):
+        infer.seed_argument(name, index, ty)
 
     # Seed return type
     if return_type is not None:
@@ -676,7 +697,7 @@ def type_inference_stage(typingctx, interp, args, return_type, locals={}):
 def native_lowering_stage(targetctx, library, interp, typemap, restype,
                           calltypes, flags):
     # Lowering
-    fndesc = lowering.PythonFunctionDescriptor.from_specialized_function(
+    fndesc = funcdesc.PythonFunctionDescriptor.from_specialized_function(
         interp, typemap, restype, calltypes, mangler=targetctx.mangler,
         inline=flags.forceinline)
 
@@ -685,36 +706,36 @@ def native_lowering_stage(targetctx, library, interp, typemap, restype,
     if not flags.no_cpython_wrapper:
         lower.create_cpython_wrapper(flags.release_gil)
     env = lower.env
-    exception_map = lower.exceptions
+    call_helper = lower.call_helper
     del lower
 
     if flags.no_compile:
-        return _LowerResult(fndesc, exception_map, cfunc=None, env=None)
+        return _LowerResult(fndesc, call_helper, cfunc=None, env=None)
     else:
         # Prepare for execution
         cfunc = targetctx.get_executable(library, fndesc, env)
         # Insert native function for use by other jitted-functions.
         # We also register its library to allow for inlining.
         targetctx.insert_user_function(cfunc, fndesc, [library])
-        return _LowerResult(fndesc, exception_map, cfunc=cfunc, env=None)
+        return _LowerResult(fndesc, call_helper, cfunc=cfunc, env=None)
 
 
 def py_lowering_stage(targetctx, library, interp, flags):
-    fndesc = lowering.PythonFunctionDescriptor.from_object_mode_function(interp)
+    fndesc = funcdesc.PythonFunctionDescriptor.from_object_mode_function(interp)
     lower = objmode.PyLower(targetctx, library, fndesc, interp)
     lower.lower()
     if not flags.no_cpython_wrapper:
         lower.create_cpython_wrapper()
     env = lower.env
-    exception_map = lower.exceptions
+    call_helper = lower.call_helper
     del lower
 
     if flags.no_compile:
-        return _LowerResult(fndesc, exception_map, cfunc=None, env=env)
+        return _LowerResult(fndesc, call_helper, cfunc=None, env=env)
     else:
         # Prepare for execution
         cfunc = targetctx.get_executable(library, fndesc, env)
-        return _LowerResult(fndesc, exception_map, cfunc, env)
+        return _LowerResult(fndesc, call_helper, cfunc, env)
 
 
 def ir_optimize_for_py_stage(interp):
