@@ -6,6 +6,7 @@ import llvmlite.llvmpy.core as lc
 import llvmlite.llvmpy.ee as le
 import llvmlite.binding as ll
 from numba.targets.imputils import implement, Registry
+from numba.targets.npyimpl import register_casters
 from numba import cgutils
 from numba import types
 from .cudadrv import nvvm
@@ -14,6 +15,7 @@ from . import nvvmutils, stubs
 registry = Registry()
 register = registry.register
 
+register_casters(register)
 
 @register
 @implement('ptx.grid.1d', types.intp)
@@ -143,13 +145,14 @@ def ptx_cmem_arylike(context, builder, sig, args):
 
     # Create array object
     ary = context.make_array(aryty)(context, builder)
-    ary.data = builder.bitcast(genptr, ary.data.type)
-
     kshape = [context.get_constant(types.intp, s) for s in arr.shape]
     kstrides = [context.get_constant(types.intp, s) for s in arr.strides]
-
-    ary.shape = cgutils.pack_array(builder, kshape)
-    ary.strides = cgutils.pack_array(builder, kstrides)
+    context.populate_array(ary,
+                           data=builder.bitcast(genptr, ary.data.type),
+                           shape=cgutils.pack_array(builder, kshape),
+                           strides=cgutils.pack_array(builder, kstrides),
+                           itemsize=ary.itemsize,
+                           parent=ary.parent)
 
     return ary._getvalue()
 
@@ -237,6 +240,9 @@ def ptx_atomic_add_intp(context, builder, sig, args):
     if aryty.dtype == types.float32:
         lmod = cgutils.get_module(builder)
         return builder.call(nvvmutils.declare_atomic_add_float32(lmod), (ptr, val))
+    elif aryty.dtype == types.float64:
+        lmod = cgutils.get_module(builder)
+        return builder.call(nvvmutils.declare_atomic_add_float64(lmod), (ptr, val))
     else:
         return builder.atomic_rmw('add', ptr, val, 'monotonic')
 
@@ -244,33 +250,6 @@ def ptx_atomic_add_intp(context, builder, sig, args):
 @register
 @implement(stubs.atomic.add, types.Kind(types.Array),
            types.Kind(types.UniTuple), types.Any)
-def ptx_atomic_add_unituple(context, builder, sig, args):
-    aryty, indty, valty = sig.args
-    ary, inds, val = args
-    dtype = aryty.dtype
-
-    indices = cgutils.unpack_tuple(builder, inds, count=len(indty))
-    indices = [context.cast(builder, i, t, types.intp)
-               for t, i in zip(indty, indices)]
-
-    if dtype != valty:
-        raise TypeError("expect %s but got %s" % (dtype, valty))
-
-    if aryty.ndim != len(indty):
-        raise TypeError("indexing %d-D array with %d-D index" %
-                        (aryty.ndim, len(indty)))
-
-    lary = context.make_array(aryty)(context, builder, ary)
-    ptr = cgutils.get_item_pointer(builder, aryty, lary, indices)
-
-    if aryty.dtype == types.float32:
-        lmod = cgutils.get_module(builder)
-        return builder.call(nvvmutils.declare_atomic_add_float32(lmod), (ptr, val))
-    else:
-        return builder.atomic_rmw('add', ptr, val, 'monotonic')
-
-
-@register
 @implement(stubs.atomic.add, types.Kind(types.Array),
            types.Kind(types.Tuple), types.Any)
 def ptx_atomic_add_tuple(context, builder, sig, args):
@@ -295,10 +274,67 @@ def ptx_atomic_add_tuple(context, builder, sig, args):
     if aryty.dtype == types.float32:
         lmod = cgutils.get_module(builder)
         return builder.call(nvvmutils.declare_atomic_add_float32(lmod), (ptr, val))
+    elif aryty.dtype == types.float64:
+        lmod = cgutils.get_module(builder)
+        return builder.call(nvvmutils.declare_atomic_add_float64(lmod), (ptr, val))
     else:
         return builder.atomic_rmw('add', ptr, val, 'monotonic')
 
 
+@register
+@implement(stubs.atomic.max, types.Kind(types.Array), types.intp, types.Any)
+def ptx_atomic_max_intp(context, builder, sig, args):
+    aryty, indty, valty = sig.args
+    ary, ind, val = args
+    dtype = aryty.dtype
+
+    if dtype != valty:
+        raise TypeError("expect %s but got %s" % (dtype, valty))
+    if aryty.ndim != 1:
+        raise TypeError("indexing %d-D array with 1-D index" % (aryty.ndim,))
+
+    lary = context.make_array(aryty)(context, builder, ary)
+    ptr = cgutils.get_item_pointer(builder, aryty, lary, [ind])
+
+    if dtype == types.float64:
+        lmod = cgutils.get_module(builder)
+        return builder.call(nvvmutils.declare_atomic_max_float64(lmod), (ptr, val))
+    else:
+        raise TypeError('Unimplemented atomic max with %s array' % dtype)
+
+
+@register
+@implement(stubs.atomic.max, types.Kind(types.Array),
+           types.Kind(types.Tuple), types.Any)
+@implement(stubs.atomic.max, types.Kind(types.Array),
+           types.Kind(types.UniTuple), types.Any)
+def ptx_atomic_max_tuple(context, builder, sig, args):
+    aryty, indty, valty = sig.args
+    ary, inds, val = args
+    dtype = aryty.dtype
+
+    indices = cgutils.unpack_tuple(builder, inds, count=len(indty))
+    indices = [context.cast(builder, i, t, types.intp)
+               for t, i in zip(indty, indices)]
+
+    if dtype != valty:
+        raise TypeError("expect %s but got %s" % (dtype, valty))
+
+    if aryty.ndim != len(indty):
+        raise TypeError("indexing %d-D array with %d-D index" %
+                        (aryty.ndim, len(indty)))
+
+    lary = context.make_array(aryty)(context, builder, ary)
+    ptr = cgutils.get_item_pointer(builder, aryty, lary, indices)
+
+    if aryty.dtype == types.float64:
+        lmod = cgutils.get_module(builder)
+        return builder.call(nvvmutils.declare_atomic_max_float64(lmod), (ptr, val))
+    else:
+        raise TypeError('Unimplemented atomic max with %s array' % dtype)
+
+
+
 # -----------------------------------------------------------------------------
 
 
@@ -353,7 +389,6 @@ def _make_array(context, builder, dataptr, dtype, shape, layout='C'):
     # Create array object
     aryty = types.Array(dtype=dtype, ndim=ndim, layout='C')
     ary = context.make_array(aryty)(context, builder)
-    ary.data = builder.bitcast(dataptr, ary.data.type)
 
     targetdata = _get_target_data(context)
     lldtype = context.get_data_type(dtype)
@@ -367,7 +402,9 @@ def _make_array(context, builder, dataptr, dtype, shape, layout='C'):
     kshape = [context.get_constant(types.intp, s) for s in shape]
     kstrides = [context.get_constant(types.intp, s) for s in strides]
 
-    ary.shape = cgutils.pack_array(builder, kshape)
-    ary.strides = cgutils.pack_array(builder, kstrides)
-
+    context.populate_array(ary,
+                           data=builder.bitcast(dataptr, ary.data.type),
+                           shape=cgutils.pack_array(builder, kshape),
+                           strides = cgutils.pack_array(builder, kstrides),
+                           itemsize=context.get_constant(types.intp, itemsize))
     return ary._getvalue()
