@@ -20,7 +20,6 @@ class DataFlowAnalysis(object):
         self.bytecode = cfa.bytecode
         # { block offset -> BlockInfo }
         self.infos = {}
-        self.syntax_blocks = []
 
     def run(self):
         for blk in self.cfa.iterliveblocks():
@@ -52,9 +51,18 @@ class DataFlowAnalysis(object):
                 warnings.warn("inconsistent stack offset for %s" % blk,
                               RuntimeWarning)
 
+            # Compute syntax blocks at block entry
+            assert ib.syntax_blocks is not None, ib
+            if info.syntax_blocks is None:
+                info.syntax_blocks = ib.syntax_blocks[:]
+            elif info.syntax_blocks != ib.syntax_blocks:
+                warnings.warn("inconsistent entry syntax blocks for %s" % blk,
+                              RuntimeWarning)
+
         if info.stack_offset is None:
             # No incoming blocks => assume it's the entry block
             info.stack_offset = 0
+            info.syntax_blocks = []
         info.stack_effect = 0
 
         for offset in blk:
@@ -63,7 +71,7 @@ class DataFlowAnalysis(object):
         return info
 
     def dump(self):
-        for blk in utils.dict_itervalues(self.infos):
+        for blk in utils.itervalues(self.infos):
             blk.dump()
 
     def dispatch(self, info, inst):
@@ -88,13 +96,13 @@ class DataFlowAnalysis(object):
         Add an inner syntax block.
         """
         block.stack_offset = info.stack_offset
-        self.syntax_blocks.append(block)
+        info.syntax_blocks.append(block)
 
     def pop_syntax_block(self, info):
         """
         Pop the innermost syntax block and revert its stack effect.
         """
-        block = self.syntax_blocks.pop()
+        block = info.syntax_blocks.pop()
         assert info.stack_offset >= block.stack_offset
         while info.stack_offset + info.stack_effect > block.stack_offset:
             info.pop(discard=True)
@@ -178,6 +186,10 @@ class DataFlowAnalysis(object):
         value = info.pop()
         info.append(inst, target=target, value=value)
 
+    def op_DELETE_ATTR(self, info, inst):
+        target = info.pop()
+        info.append(inst, target=target)
+
     def op_STORE_FAST(self, info, inst):
         value = info.pop()
         info.append(inst, value=value)
@@ -233,14 +245,9 @@ class DataFlowAnalysis(object):
         res = info.make_temp()
         info.append(inst, value=value, res=res)
         info.push(res)
-        if self.syntax_blocks:
-            loop = self.syntax_blocks[-1]
-            if isinstance(loop, LoopBlock) and loop.iterator is None:
-                loop.iterator = res
 
     def op_FOR_ITER(self, info, inst):
-        loop = self.syntax_blocks[-1]
-        iterator = loop.iterator
+        iterator = info.tos
         pair = info.make_temp()
         indval = info.make_temp()
         pred = info.make_temp()
@@ -491,28 +498,35 @@ class DataFlowAnalysis(object):
         info.terminator = inst
 
     def op_BREAK_LOOP(self, info, inst):
+        self.pop_syntax_block(info)
         info.append(inst)
         info.terminator = inst
 
     def op_RETURN_VALUE(self, info, inst):
-        info.append(inst, retval=info.pop())
+        info.append(inst, retval=info.pop(), castval=info.make_temp())
         info.terminator = inst
 
+    def op_YIELD_VALUE(self, info, inst):
+        val = info.pop()
+        res = info.make_temp()
+        info.append(inst, value=val, res=res)
+        info.push(res)
+
     def op_SETUP_LOOP(self, info, inst):
         self.add_syntax_block(info, LoopBlock())
         info.append(inst)
 
     def op_POP_BLOCK(self, info, inst):
         block = self.pop_syntax_block(info)
-        if isinstance(block, LoopBlock):
-            info.append(inst, delitems=[block.iterator])
-        else:
-            info.append(inst)
+        info.append(inst)
 
     def op_RAISE_VARARGS(self, info, inst):
-        if inst.arg != 1:
+        if inst.arg == 0:
+            exc = None
+        elif inst.arg == 1:
+            exc = info.pop()
+        else:
             raise ValueError("Multiple argument raise is not supported.")
-        exc = info.pop()
         info.append(inst, exc=exc)
 
     def _ignored(self, info, inst):
@@ -520,10 +534,9 @@ class DataFlowAnalysis(object):
 
 
 class LoopBlock(object):
-    __slots__ = ('iterator', 'stack_offset')
+    __slots__ = ('stack_offset',)
 
     def __init__(self):
-        self.iterator = None
         self.stack_offset = None
 
 
@@ -542,6 +555,7 @@ class BlockInfo(object):
         self._term = None
         self.stack_offset = None
         self.stack_effect = 0
+        self.syntax_blocks = None
 
     def __repr__(self):
         return "<%s at offset %d>" % (self.__class__.__name__, self.offset)
@@ -555,7 +569,7 @@ class BlockInfo(object):
 
     def make_temp(self, prefix=''):
         self.tempct += 1
-        name = '$%s%d.%d' % (prefix, self.offset, self.tempct)
+        name = '$%s%s.%s' % (prefix, self.offset, self.tempct)
         return name
 
     def push(self, val):
