@@ -1,6 +1,5 @@
 from __future__ import print_function, absolute_import, division
 
-
 import math
 from functools import reduce
 
@@ -10,7 +9,7 @@ import llvmlite.llvmpy.core as lc
 from .imputils import (builtin, builtin_attr, implement, impl_attribute,
                        iternext_impl, struct_factory)
 from . import optional
-from .. import typing, types, cgutils, utils, errcode, intrinsics
+from .. import typing, types, cgutils, utils, intrinsics
 
 #-------------------------------------------------------------------------------
 
@@ -44,7 +43,8 @@ def int_udiv_impl(context, builder, sig, args):
     [ta, tb] = sig.args
     a = context.cast(builder, va, ta, sig.return_type)
     b = context.cast(builder, vb, tb, sig.return_type)
-    cgutils.guard_zero(context, builder, b)
+    cgutils.guard_zero(context, builder, b,
+                       (ZeroDivisionError, "integer division by zero"))
     return builder.udiv(a, b)
 
 
@@ -99,7 +99,8 @@ def int_sdiv_impl(context, builder, sig, args):
     [ta, tb] = sig.args
     a = context.cast(builder, va, ta, sig.return_type)
     b = context.cast(builder, vb, tb, sig.return_type)
-    cgutils.guard_zero(context, builder, b)
+    cgutils.guard_zero(context, builder, b,
+                       (ZeroDivisionError, "integer division by zero"))
     div, _ = int_divmod(context, builder, a, b)
     return div
 
@@ -108,7 +109,8 @@ def int_struediv_impl(context, builder, sig, args):
     x, y = args
     fx = builder.sitofp(x, Type.double())
     fy = builder.sitofp(y, Type.double())
-    cgutils.guard_zero(context, builder, y)
+    cgutils.guard_zero(context, builder, y,
+                       (ZeroDivisionError, "division by zero"))
     return builder.fdiv(fx, fy)
 
 
@@ -116,7 +118,8 @@ def int_utruediv_impl(context, builder, sig, args):
     x, y = args
     fx = builder.uitofp(x, Type.double())
     fy = builder.uitofp(y, Type.double())
-    cgutils.guard_zero(context, builder, y)
+    cgutils.guard_zero(context, builder, y,
+                       (ZeroDivisionError, "division by zero"))
     return builder.fdiv(fx, fy)
 
 
@@ -126,13 +129,16 @@ int_ufloordiv_impl = int_udiv_impl
 
 def int_srem_impl(context, builder, sig, args):
     x, y = args
-    cgutils.guard_zero(context, builder, y)
+    cgutils.guard_zero(context, builder, y,
+                       (ZeroDivisionError, "integer modulo by zero"))
     _, rem = int_divmod(context, builder, x, y)
     return rem
 
 
 def int_urem_impl(context, builder, sig, args):
     x, y = args
+    cgutils.guard_zero(context, builder, y,
+                       (ZeroDivisionError, "integer modulo by zero"))
     return builder.urem(x, y)
 
 
@@ -385,7 +391,10 @@ builtin(implement('>', types.boolean, types.boolean)(int_ugt_impl))
 builtin(implement('>=', types.boolean, types.boolean)(int_uge_impl))
 builtin(implement('~', types.boolean)(bool_invert_impl))
 
-for ty in types.integer_domain:
+
+def _implement_integer_operators():
+    ty = types.Kind(types.Integer)
+
     builtin(implement('+', ty, ty)(int_add_impl))
     builtin(implement('-', ty, ty)(int_sub_impl))
     builtin(implement('*', ty, ty)(int_mul_impl))
@@ -404,38 +413,41 @@ for ty in types.integer_domain:
     builtin(implement('~', ty)(int_invert_impl))
     builtin(implement(types.sign_type, ty)(int_sign_impl))
 
-for ty in types.unsigned_domain:
-    builtin(implement('/?', ty, ty)(int_udiv_impl))
-    builtin(implement('//', ty, ty)(int_ufloordiv_impl))
-    builtin(implement('/', ty, ty)(int_utruediv_impl))
-    builtin(implement('%', ty, ty)(int_urem_impl))
-    builtin(implement('<', ty, ty)(int_ult_impl))
-    builtin(implement('<=', ty, ty)(int_ule_impl))
-    builtin(implement('>', ty, ty)(int_ugt_impl))
-    builtin(implement('>=', ty, ty)(int_uge_impl))
-    builtin(implement('**', types.float64, ty)(int_upower_impl))
-    builtin(implement(pow, types.float64, ty)(int_upower_impl))
-    # logical shift for unsigned
-    builtin(implement('>>', ty, types.uint32)(int_lshr_impl))
-    builtin(implement(types.abs_type, ty)(uint_abs_impl))
+    for ty in types.unsigned_domain:
+        builtin(implement('/?', ty, ty)(int_udiv_impl))
+        builtin(implement('//', ty, ty)(int_ufloordiv_impl))
+        builtin(implement('/', ty, ty)(int_utruediv_impl))
+        builtin(implement('%', ty, ty)(int_urem_impl))
+        builtin(implement('<', ty, ty)(int_ult_impl))
+        builtin(implement('<=', ty, ty)(int_ule_impl))
+        builtin(implement('>', ty, ty)(int_ugt_impl))
+        builtin(implement('>=', ty, ty)(int_uge_impl))
+        builtin(implement('**', types.float64, ty)(int_upower_impl))
+        builtin(implement(pow, types.float64, ty)(int_upower_impl))
+        # logical shift for unsigned
+        builtin(implement('>>', ty, types.uint32)(int_lshr_impl))
+        builtin(implement(types.abs_type, ty)(uint_abs_impl))
 
-for ty in types.signed_domain:
-    builtin(implement('/?', ty, ty)(int_sdiv_impl))
-    builtin(implement('//', ty, ty)(int_sfloordiv_impl))
-    builtin(implement('/', ty, ty)(int_struediv_impl))
-    builtin(implement('%', ty, ty)(int_srem_impl))
-    builtin(implement('<', ty, ty)(int_slt_impl))
-    builtin(implement('<=', ty, ty)(int_sle_impl))
-    builtin(implement('>', ty, ty)(int_sgt_impl))
-    builtin(implement('>=', ty, ty)(int_sge_impl))
-    builtin(implement(types.abs_type, ty)(int_abs_impl))
-    builtin(implement('**', types.float64, ty)(int_spower_impl))
-    builtin(implement(pow, types.float64, ty)(int_spower_impl))
-    # arithmetic shift for signed
-    builtin(implement('>>', ty, types.uint32)(int_ashr_impl))
+    for ty in types.signed_domain:
+        builtin(implement('/?', ty, ty)(int_sdiv_impl))
+        builtin(implement('//', ty, ty)(int_sfloordiv_impl))
+        builtin(implement('/', ty, ty)(int_struediv_impl))
+        builtin(implement('%', ty, ty)(int_srem_impl))
+        builtin(implement('<', ty, ty)(int_slt_impl))
+        builtin(implement('<=', ty, ty)(int_sle_impl))
+        builtin(implement('>', ty, ty)(int_sgt_impl))
+        builtin(implement('>=', ty, ty)(int_sge_impl))
+        builtin(implement(types.abs_type, ty)(int_abs_impl))
+        builtin(implement('**', types.float64, ty)(int_spower_impl))
+        builtin(implement(pow, types.float64, ty)(int_spower_impl))
+        # arithmetic shift for signed
+        builtin(implement('>>', ty, types.uint32)(int_ashr_impl))
+
+    builtin(implement('**', types.intp, types.intp)(int_spower_impl))
+    builtin(implement(pow, types.intp, types.intp)(int_spower_impl))
+
+_implement_integer_operators()
 
-builtin(implement('**', types.intp, types.intp)(int_spower_impl))
-builtin(implement(pow, types.intp, types.intp)(int_spower_impl))
 
 def optional_is_none(context, builder, sig, args):
     """Check if an Optional value is invalid
@@ -498,7 +510,8 @@ def real_mul_impl(context, builder, sig, args):
 
 
 def real_div_impl(context, builder, sig, args):
-    cgutils.guard_zero(context, builder, args[1])
+    cgutils.guard_zero(context, builder, args[1],
+                       (ZeroDivisionError, "division by zero"))
     return builder.fdiv(*args)
 
 
@@ -635,14 +648,16 @@ def real_divmod_func_body(context, builder, vx, wx):
 
 def real_mod_impl(context, builder, sig, args):
     x, y = args
-    cgutils.guard_zero(context, builder, y)
+    cgutils.guard_zero(context, builder, args[1],
+                       (ZeroDivisionError, "modulo by zero"))
     _, rem = real_divmod(context, builder, x, y)
     return rem
 
 
 def real_floordiv_impl(context, builder, sig, args):
     x, y = args
-    cgutils.guard_zero(context, builder, y)
+    cgutils.guard_zero(context, builder, args[1],
+                       (ZeroDivisionError, "division by zero"))
     quot, _ = real_divmod(context, builder, x, y)
     return quot
 
@@ -724,30 +739,33 @@ def real_sign_impl(context, builder, sig, args):
     return builder.load(presult)
 
 
-for ty in types.real_domain:
-    builtin(implement('+', ty, ty)(real_add_impl))
-    builtin(implement('-', ty, ty)(real_sub_impl))
-    builtin(implement('*', ty, ty)(real_mul_impl))
-    builtin(implement('/?', ty, ty)(real_div_impl))
-    builtin(implement('//', ty, ty)(real_floordiv_impl))
-    builtin(implement('/', ty, ty)(real_div_impl))
-    builtin(implement('%', ty, ty)(real_mod_impl))
-    builtin(implement('**', ty, ty)(real_power_impl))
-    builtin(implement(pow, ty, ty)(real_power_impl))
+ty = types.Kind(types.Float)
 
-    builtin(implement('==', ty, ty)(real_eq_impl))
-    builtin(implement('!=', ty, ty)(real_ne_impl))
-    builtin(implement('<', ty, ty)(real_lt_impl))
-    builtin(implement('<=', ty, ty)(real_le_impl))
-    builtin(implement('>', ty, ty)(real_gt_impl))
-    builtin(implement('>=', ty, ty)(real_ge_impl))
+builtin(implement('+', ty, ty)(real_add_impl))
+builtin(implement('-', ty, ty)(real_sub_impl))
+builtin(implement('*', ty, ty)(real_mul_impl))
+builtin(implement('/?', ty, ty)(real_div_impl))
+builtin(implement('//', ty, ty)(real_floordiv_impl))
+builtin(implement('/', ty, ty)(real_div_impl))
+builtin(implement('%', ty, ty)(real_mod_impl))
+builtin(implement('**', ty, ty)(real_power_impl))
+builtin(implement(pow, ty, ty)(real_power_impl))
 
-    builtin(implement(types.abs_type, ty)(real_abs_impl))
+builtin(implement('==', ty, ty)(real_eq_impl))
+builtin(implement('!=', ty, ty)(real_ne_impl))
+builtin(implement('<', ty, ty)(real_lt_impl))
+builtin(implement('<=', ty, ty)(real_le_impl))
+builtin(implement('>', ty, ty)(real_gt_impl))
+builtin(implement('>=', ty, ty)(real_ge_impl))
 
-    builtin(implement('-', ty)(real_negate_impl))
-    builtin(implement('+', ty)(real_positive_impl))
-    builtin(implement(types.neg_type, ty)(real_negate_impl))
-    builtin(implement(types.sign_type, ty)(real_sign_impl))
+builtin(implement(types.abs_type, ty)(real_abs_impl))
+
+builtin(implement('-', ty)(real_negate_impl))
+builtin(implement('+', ty)(real_positive_impl))
+builtin(implement(types.neg_type, ty)(real_negate_impl))
+builtin(implement(types.sign_type, ty)(real_sign_impl))
+
+del ty
 
 
 class Complex64(cgutils.Structure):
@@ -979,20 +997,23 @@ def complex_abs_impl(context, builder, sig, args):
     return context.compile_internal(builder, complex_abs, sig, args)
 
 
-for ty in types.complex_domain:
-    builtin(implement("+", ty, ty)(complex_add_impl))
-    builtin(implement("-", ty, ty)(complex_sub_impl))
-    builtin(implement("*", ty, ty)(complex_mul_impl))
-    builtin(implement("/?", ty, ty)(complex_div_impl))
-    builtin(implement("/", ty, ty)(complex_div_impl))
-    builtin(implement("-", ty)(complex_negate_impl))
-    builtin(implement("+", ty)(complex_positive_impl))
-    # Complex modulo is deprecated in python3
+ty = types.Kind(types.Complex)
 
-    builtin(implement('==', ty, ty)(complex_eq_impl))
-    builtin(implement('!=', ty, ty)(complex_ne_impl))
+builtin(implement("+", ty, ty)(complex_add_impl))
+builtin(implement("-", ty, ty)(complex_sub_impl))
+builtin(implement("*", ty, ty)(complex_mul_impl))
+builtin(implement("/?", ty, ty)(complex_div_impl))
+builtin(implement("/", ty, ty)(complex_div_impl))
+builtin(implement("-", ty)(complex_negate_impl))
+builtin(implement("+", ty)(complex_positive_impl))
+# Complex modulo is deprecated in python3
 
-    builtin(implement(types.abs_type, ty)(complex_abs_impl))
+builtin(implement('==', ty, ty)(complex_eq_impl))
+builtin(implement('!=', ty, ty)(complex_ne_impl))
+
+builtin(implement(types.abs_type, ty)(complex_abs_impl))
+
+del ty
 
 
 #------------------------------------------------------------------------------
@@ -1012,9 +1033,9 @@ def number_as_bool_impl(context, builder, sig, args):
     return istrue
 
 
-for ty in types.number_domain:
-    builtin(implement('not', ty)(number_not_impl))
-    builtin(implement(bool, ty)(number_as_bool_impl))
+for ty in (types.Integer, types.Float, types.Complex):
+    builtin(implement('not', types.Kind(ty))(number_not_impl))
+    builtin(implement(bool, types.Kind(ty))(number_as_bool_impl))
 
 builtin(implement('not', types.boolean)(number_not_impl))
 
@@ -1102,11 +1123,7 @@ def slice0_none_none_impl(context, builder, sig, args):
 
 
 def make_pair(first_type, second_type):
-    class Pair(cgutils.Structure):
-        _fields = [('first', first_type),
-                   ('second', second_type)]
-
-    return Pair
+    return cgutils.create_struct_proxy(types.Pair(first_type, second_type))
 
 
 @struct_factory(types.UniTupleIter)
@@ -1115,13 +1132,7 @@ def make_unituple_iter(tupiter):
     Return the Structure representation of the given *tupiter* (an
     instance of types.UniTupleIter).
     """
-    unituple = tupiter.unituple
-
-    class UniTupleIter(cgutils.Structure):
-        _fields = [('index', types.CPointer(types.intp)),
-                   ('tuple', tupiter.unituple,)]
-
-    return UniTupleIter
+    return cgutils.create_struct_proxy(tupiter)
 
 
 @builtin
@@ -1183,7 +1194,8 @@ def getitem_unituple(context, builder, sig, args):
     switch = builder.switch(idx, bbelse, n=tupty.count)
 
     with cgutils.goto_block(builder, bbelse):
-        context.return_errcode(builder, errcode.OUT_OF_BOUND_ERROR)
+        context.call_conv.return_user_exc(builder, IndexError,
+                                          ("tuple index out of range",))
 
     lrtty = context.get_value_type(tupty.dtype)
     with cgutils.goto_block(builder, bbend):
@@ -1269,30 +1281,62 @@ def min_impl(context, builder, sig, args):
     return resval
 
 
-@builtin
-@implement(round, types.float32)
-def round_impl_f32(context, builder, sig, args):
-    module = cgutils.get_module(builder)
-    fnty = Type.function(Type.float(), [Type.float()])
+def _round_intrinsic(tp):
+    # round() rounds half to even on Python 3, away from zero on Python 2.
     if utils.IS_PY3:
-        fn = module.get_or_insert_function(fnty, name="numba.roundf")
+        return "llvm.rint.f%d" % (tp.bitwidth,)
     else:
-        fn = module.get_or_insert_function(fnty, name="roundf")
-    assert fn.is_declaration
-    return builder.call(fn, args)
-
+        return "llvm.round.f%d" % (tp.bitwidth,)
 
 @builtin
-@implement(round, types.float64)
-def round_impl_f64(context, builder, sig, args):
+@implement(round, types.Kind(types.Float))
+def round_impl_unary(context, builder, sig, args):
+    fltty = sig.args[0]
+    llty = context.get_value_type(fltty)
     module = cgutils.get_module(builder)
-    fnty = Type.function(Type.double(), [Type.double()])
+    fnty = Type.function(llty, [llty])
+    fn = module.get_or_insert_function(fnty, name=_round_intrinsic(fltty))
+    res = builder.call(fn, args)
     if utils.IS_PY3:
-        fn = module.get_or_insert_function(fnty, name="numba.round")
+        # unary round() returns an int on Python 3
+        return builder.fptosi(res, context.get_value_type(sig.return_type))
     else:
-        fn = module.get_or_insert_function(fnty, name="round")
-    assert fn.is_declaration
-    return builder.call(fn, args)
+        return res
+
+@builtin
+@implement(round, types.Kind(types.Float), types.Kind(types.Integer))
+def round_impl_binary(context, builder, sig, args):
+    fltty = sig.args[0]
+    # Allow calling the intrinsic from the Python implementation below.
+    # This avoids the conversion to an int in Python 3's unary round().
+    _round = types.ExternalFunction(
+        _round_intrinsic(fltty), typing.signature(fltty, fltty))
+
+    def round_ndigits(x, ndigits):
+        if math.isinf(x) or math.isnan(x):
+            return x
+
+        if ndigits >= 0:
+            if ndigits > 22:
+                # pow1 and pow2 are each safe from overflow, but
+                # pow1*pow2 ~= pow(10.0, ndigits) might overflow.
+                pow1 = 10.0 ** (ndigits - 22)
+                pow2 = 1e22
+            else:
+                pow1 = 10.0 ** ndigits
+                pow2 = 1.0
+            y = (x * pow1) * pow2
+            if math.isinf(y):
+                return x
+            return (_round(y) / pow2) / pow1
+
+        else:
+            pow1 = 10.0 ** (-ndigits)
+            y = x / pow1
+            return _round(y) * pow1
+
+    return context.compile_internal(builder, round_ndigits, sig, args)
+
 
 #-------------------------------------------------------------------------------
 
@@ -1372,9 +1416,28 @@ def array_ravel_impl(context, builder, sig, args):
     strides = cgutils.unpack_tuple(builder, arr.strides, arrty.ndim)
     unit_stride = strides[0] if arrty.layout == 'F' else strides[-1]
 
-    flatarr.data = arr.data
-    flatarr.shape = cgutils.pack_array(builder, [size])
-    flatarr.strides = cgutils.pack_array(builder, [unit_stride])
+    context.populate_array(flatarr,
+                           data=arr.data,
+                           shape=cgutils.pack_array(builder, [size]),
+                           strides=cgutils.pack_array(builder, [unit_stride]),
+                           itemsize=arr.itemsize,
+                           parent=arr.parent)
 
     return flatarr._getvalue()
 
+
+# -----------------------------------------------------------------------------
+
+@builtin
+@implement(types.len_type, types.Kind(types.Tuple))
+def array_len(context, builder, sig, args):
+    tupty, = sig.args
+    retty = sig.return_type
+    return context.get_constant(retty, len(tupty.types))
+
+@builtin
+@implement(types.len_type, types.Kind(types.UniTuple))
+def array_len(context, builder, sig, args):
+    tupty, = sig.args
+    retty = sig.return_type
+    return context.get_constant(retty, tupty.count)
