@@ -1,5 +1,6 @@
 from __future__ import print_function, division, absolute_import
 
+import functools
 import sys
 
 import llvmlite.llvmpy.core as lc
@@ -7,7 +8,7 @@ import llvmlite.llvmpy.passes as lp
 import llvmlite.binding as ll
 import llvmlite.ir as llvmir
 
-from numba import config
+from numba import config, utils
 
 
 _x86arch = frozenset(['x86', 'i386', 'i486', 'i586', 'i686', 'i786',
@@ -18,6 +19,12 @@ def _is_x86(triple):
     return arch in _x86arch
 
 
+def dump(header, body):
+    print(header.center(80, '-'))
+    print(body)
+    print('=' * 80)
+
+
 class CodeLibrary(object):
     """
     An interface for bundling LLVM code together and compiling it.
@@ -135,9 +142,7 @@ class CodeLibrary(object):
         self._raise_if_finalized()
 
         if config.DUMP_FUNC_OPT:
-            print(("FUNCTION OPTIMIZED DUMP %s" % self._name).center(80, '-'))
-            print(self._final_module)
-            print('=' * 80)
+            dump("FUNCTION OPTIMIZED DUMP %s" % self._name, self.get_llvm_str())
 
         # Link libraries for shared code
         for library in self._linking_libraries:
@@ -155,29 +160,37 @@ class CodeLibrary(object):
         # It seems add_module() must be done only here and not before
         # linking in other modules, otherwise get_pointer_to_function()
         # could fail.
-        self._codegen._add_module(self._final_module)
+        cleanup = self._codegen._add_module(self._final_module)
+        if cleanup:
+            utils.finalize(self, cleanup)
         self._finalize_specific()
 
         self._finalized = True
 
         if config.DUMP_OPTIMIZED:
-            print(("OPTIMIZED DUMP %s" % self._name).center(80, '-'))
-            print(self._final_module)
-            print('=' * 80)
+            dump("OPTIMIZED DUMP %s" % self._name, self.get_llvm_str())
 
         if config.DUMP_ASSEMBLY:
-            self._dump_assembly()
+            # CUDA backend cannot return assembly this early, so don't
+            # attempt to dump assembly if nothing is produced.
+            asm = self.get_asm_str()
+            if asm:
+                dump("ASSEMBLY %s" % self._name, self.get_asm_str())
 
     def get_function(self, name):
         return self._final_module.get_function(name)
 
-    def _dump_assembly(self):
+    def get_llvm_str(self):
         """
-        Internal: dump native assembler code for this library.
+        Get the human-readable form of the LLVM module.
         """
-        print(("ASSEMBLY %s" % self._name).center(80, '-'))
-        print(self._codegen._tm.emit_assembly(self._final_module))
-        print('=' * 80)
+        return str(self._final_module)
+
+    def get_asm_str(self):
+        """
+        Get the human-readable assembly.
+        """
+        return str(self._codegen._tm.emit_assembly(self._final_module))
 
 
 class AOTCodeLibrary(CodeLibrary):
@@ -363,3 +376,5 @@ class JITCPUCodegen(BaseCPUCodegen):
 
     def _add_module(self, module):
         self._engine.add_module(module)
+        # Early bind the engine method to avoid keeping a reference to self.
+        return functools.partial(self._engine.remove_module, module)
