@@ -10,8 +10,10 @@ from numba import _dynfunc, config
 from numba.callwrapper import PyCallWrapper
 from .base import BaseContext, PYOBJECT
 from numba import utils, cgutils, types
+from numba.utils import cached_property
 from numba.targets import (
-    codegen, externals, intrinsics, cmathimpl, mathimpl, npyimpl, operatorimpl, printimpl)
+    callconv, codegen, externals, intrinsics, cmathimpl, mathimpl,
+    npyimpl, operatorimpl, printimpl, randomimpl)
 from .options import TargetOptions
 
 
@@ -37,7 +39,6 @@ class CPUContext(BaseContext):
         return self._internal_codegen._create_empty_module(name)
 
     def init(self):
-        self.native_funcs = utils.UniqueDict()
         self.is32bit = (utils.MACHINE_BITS == 32)
 
         # Map external C functions.
@@ -45,11 +46,12 @@ class CPUContext(BaseContext):
         externals.c_numpy_functions.install()
 
         # Add target specific implementations
-        self.insert_func_defn(cmathimpl.registry.functions)
-        self.insert_func_defn(mathimpl.registry.functions)
-        self.insert_func_defn(npyimpl.registry.functions)
-        self.insert_func_defn(operatorimpl.registry.functions)
-        self.insert_func_defn(printimpl.registry.functions)
+        self.install_registry(cmathimpl.registry)
+        self.install_registry(mathimpl.registry)
+        self.install_registry(npyimpl.registry)
+        self.install_registry(operatorimpl.registry)
+        self.install_registry(printimpl.registry)
+        self.install_registry(randomimpl.registry)
 
         self._internal_codegen = codegen.JITCPUCodegen("numba.exec")
 
@@ -63,105 +65,9 @@ class CPUContext(BaseContext):
     def jit_codegen(self):
         return self._internal_codegen
 
-    def get_function_type(self, fndesc):
-        """
-        Get the implemented Function type for the high-level *fndesc*.
-        Some parameters can be added or shuffled around.
-        This is kept in sync with call_function() and get_arguments().
-
-        Calling Convention
-        ------------------
-        (Same return value convention as BaseContext target.)
-        Returns: -2 for return none in native function;
-                 -1 for failure with python exception set;
-                  0 for success;
-                 >0 for user error code.
-        Return value is passed by reference as the first argument.
-
-        The 2nd argument is a _dynfunc.Environment object.
-        It MUST NOT be used if the function is in nopython mode.
-
-        Actual arguments starts at the 3rd argument position.
-        Caller is responsible to allocate space for return value.
-        """
-        return self.get_function_type2(fndesc.restype, fndesc.argtypes)
-
-    def get_function_type2(self, restype, argtypes):
-        """
-        Get the implemented Function type for the high-level *fndesc*.
-        Some parameters can be added or shuffled around.
-        This is kept in sync with call_function() and get_arguments().
-
-        Calling Convention
-        ------------------
-        (Same return value convention as BaseContext target.)
-        Returns: -2 for return none in native function;
-                 -1 for failure with python exception set;
-                  0 for success;
-                 >0 for user error code.
-        Return value is passed by reference as the first argument.
-
-        The 2nd argument is a _dynfunc.Environment object.
-        It MUST NOT be used if the function is in nopython mode.
-
-        Actual arguments starts at the 3rd argument position.
-        Caller is responsible to allocate space for return value.
-        """
-        argtypes = [self.get_argument_type(aty)
-                    for aty in argtypes]
-        resptr = self.get_return_type(restype)
-        fnty = lc.Type.function(lc.Type.int(), [resptr, PYOBJECT] + argtypes)
-        return fnty
-
-    def declare_function(self, module, fndesc):
-        """
-        Override parent to handle get_env_argument
-        """
-        fnty = self.get_function_type(fndesc)
-        fn = module.get_or_insert_function(fnty, name=fndesc.llvm_func_name)
-        assert fn.is_declaration
-        for ak, av in zip(fndesc.args, self.get_arguments(fn)):
-            av.name = "arg.%s" % ak
-        self.get_env_argument(fn).name = "env"
-        fn.args[0].name = "ret"
-        return fn
-
-    def get_arguments(self, func):
-        """Override parent to handle enviroment argument
-        Get the Python-level arguments of LLVM *func*.
-        See get_function_type() for the calling convention.
-        """
-        return func.args[2:]
-
-    def get_env_argument(self, func):
-        """
-        Get the environment argument of LLVM *func* (which can be
-        a declaration).
-        """
-        return func.args[1]
-
-    def call_function(self, builder, callee, resty, argtys, args, env=None):
-        """
-        Call the Numba-compiled *callee*, using the same calling
-        convention as in get_function_type().
-        """
-        if env is None:
-            # This only works with functions that don't use the environment
-            # (nopython functions).
-            env = lc.Constant.null(PYOBJECT)
-        retty = callee.args[0].type.pointee
-        retvaltmp = cgutils.alloca_once(builder, retty)
-        # initialize return value to zeros
-        builder.store(lc.Constant.null(retty), retvaltmp)
-
-        args = [self.get_value_as_argument(builder, ty, arg)
-                for ty, arg in zip(argtys, args)]
-        realargs = [retvaltmp, env] + list(args)
-        code = builder.call(callee, realargs)
-        status = self.get_return_status(builder, code)
-        retval = builder.load(retvaltmp)
-        out = self.get_returned_value(builder, resty, retval)
-        return status, out
+    @cached_property
+    def call_conv(self):
+        return callconv.CPUCallConv(self)
 
     def get_env_from_closure(self, builder, clo):
         """
@@ -169,7 +75,7 @@ class CPUContext(BaseContext):
         to the enclosed _dynfunc.Environment.
         """
         clo_body_ptr = cgutils.pointer_add(
-            builder, clo, _dynfunc._impl_info['offset_closure_body'])
+            builder, clo, _dynfunc._impl_info['offsetof_closure_body'])
         clo_body = ClosureBody(self, builder, ref=clo_body_ptr, cast_ref=True)
         return clo_body.env
 
@@ -179,15 +85,17 @@ class CPUContext(BaseContext):
         get a EnvBody allowing structured access to environment fields.
         """
         body_ptr = cgutils.pointer_add(
-            builder, envptr, _dynfunc._impl_info['offset_env_body'])
+            builder, envptr, _dynfunc._impl_info['offsetof_env_body'])
         return EnvBody(self, builder, ref=body_ptr, cast_ref=True)
 
-    def remove_native_function(self, func):
+    def get_generator_state(self, builder, genptr, return_type):
         """
-        Remove internal references to nonpython mode function *func*.
-        KeyError is raised if the function isn't known to us.
+        From the given *genptr* (a pointer to a _dynfunc.Generator object),
+        get a pointer to its state area.
         """
-        del self.native_funcs[func]
+        return cgutils.pointer_add(
+            builder, genptr, _dynfunc._impl_info['offsetof_generator_state'],
+            return_type=return_type)
 
     def post_lowering(self, func):
         mod = func.module
@@ -201,13 +109,13 @@ class CPUContext(BaseContext):
             # calls to compiler-rt
             intrinsics.fix_divmod(mod)
 
-    def create_cpython_wrapper(self, library, fndesc, exceptions,
+    def create_cpython_wrapper(self, library, fndesc, call_helper,
                                release_gil=False):
         wrapper_module = self.create_module("wrapper")
-        fnty = self.get_function_type(fndesc)
+        fnty = self.call_conv.get_function_type(fndesc.restype, fndesc.argtypes)
         wrapper_callee = wrapper_module.add_function(fnty, fndesc.llvm_func_name)
         builder = PyCallWrapper(self, wrapper_module, wrapper_callee,
-                                fndesc, exceptions=exceptions,
+                                fndesc, call_helper=call_helper,
                                 release_gil=release_gil)
         builder.build()
         library.add_ir_module(wrapper_module)
@@ -225,19 +133,16 @@ class CPUContext(BaseContext):
         - env
             an execution environment (from _dynfunc)
         """
-        func = library.get_function(fndesc.llvm_func_name)
-        wrapper = library.get_function(fndesc.llvm_cpython_wrapper_name)
-
         # Code generation
-        baseptr = library.get_pointer_to_function(func.name)
-        fnptr = library.get_pointer_to_function(wrapper.name)
+        baseptr = library.get_pointer_to_function(fndesc.llvm_func_name)
+        fnptr = library.get_pointer_to_function(fndesc.llvm_cpython_wrapper_name)
 
         cfunc = _dynfunc.make_function(fndesc.lookup_module(),
                                        fndesc.qualname.split('.')[-1],
-                                       fndesc.doc, fnptr, env)
-
-        if fndesc.native:
-            self.native_funcs[cfunc] = fndesc.llvm_func_name, baseptr
+                                       fndesc.doc, fnptr, env,
+                                       # objects to keepalive with the function
+                                       (library,)
+                                       )
 
         return cfunc
 
