@@ -34,13 +34,16 @@ def caster(restype, constructor):
 
     return _cast
 
-for tp in types.number_domain:
-    register(caster(tp, getattr(numpy, str(tp))))
+def register_casters(register_function):
+    for tp in types.number_domain:
+        register_function(caster(tp, getattr(numpy, str(tp))))
 
-register(caster(types.intc, numpy.intc))
-register(caster(types.uintc, numpy.uintc))
-register(caster(types.intp, numpy.intp))
-register(caster(types.uintp, numpy.uintp))
+    register_function(caster(types.intc, numpy.intc))
+    register_function(caster(types.uintc, numpy.uintc))
+    register_function(caster(types.intp, numpy.intp))
+    register_function(caster(types.uintp, numpy.uintp))
+
+register_casters(register)
 
 ########################################################################
 
@@ -164,8 +167,7 @@ class _ArrayHelper(namedtuple('_ArrayHelper', ('context', 'builder', 'ary',
         ctx = self.context
         bld = self.builder
 
-        # maybe there should be a "get_memory_value" or "get_storage_value"
-        store_value = ctx.get_struct_member_value(bld, self.base_type, value)
+        store_value = ctx.get_value_as_data(bld, self.base_type, value)
         assert ctx.get_data_type(self.base_type) == store_value.type
 
         bld.store(store_value, self._load_effective_address(indices))
@@ -319,13 +321,12 @@ def register_unary_ufunc_kernel(ufunc, kernel):
         return numpy_ufunc_kernel(context, builder, sig, args, kernel,
                                   explicit_output=False)
 
-    register(implement(ufunc, types.Kind(types.Array),
-        types.Kind(types.Array))(unary_ufunc))
-    for ty in types.number_domain:
-        register(implement(ufunc, ty,
-            types.Kind(types.Array))(unary_ufunc))
-    for ty in types.number_domain:
-        register(implement(ufunc, ty)(unary_scalar_ufunc)) # scalar
+    _any = types.Any
+
+    # (array or scalar, out=array)
+    register(implement(ufunc, _any, types.Kind(types.Array))(unary_ufunc))
+    # (scalar, out=array)
+    register(implement(ufunc, _any)(unary_scalar_ufunc))
 
 
 def register_binary_ufunc_kernel(ufunc, kernel):
@@ -336,17 +337,12 @@ def register_binary_ufunc_kernel(ufunc, kernel):
         return numpy_ufunc_kernel(context, builder, sig, args, kernel,
                                   explicit_output=False)
 
-    register(implement(ufunc, types.Kind(types.Array), types.Kind(types.Array),
-        types.Kind(types.Array))(binary_ufunc))
-    for ty in types.number_domain:
-        register(implement(ufunc, ty, types.Kind(types.Array),
-            types.Kind(types.Array))(binary_ufunc))
-        register(implement(ufunc, types.Kind(types.Array), ty,
-            types.Kind(types.Array))(binary_ufunc))
-    for ty1, ty2 in itertools.product(types.number_domain, types.number_domain):
-        register(implement(ufunc, ty1, ty2,
-            types.Kind(types.Array))(binary_ufunc))
-        register(implement(ufunc, ty1, ty2)(binary_scalar_ufunc)) # scalar
+    _any = types.Any
+
+    # (array or scalar, array o scalar, out=array)
+    register(implement(ufunc, _any, _any, types.Kind(types.Array))(binary_ufunc))
+    # (scalar, scalar)
+    register(implement(ufunc, _any, _any)(binary_scalar_ufunc))
 
 
 ################################################################################
