@@ -52,6 +52,11 @@ class TargetOptions(object):
         if kws.pop('debug', False):
             flags.set("boundcheck")
 
+        if kws.pop('nogil', False):
+            flags.set("release_gil")
+
+        flags.set("enable_pyobject_looplift")
+
         if kws:
             # Unread options?
             raise NameError("Unrecognized options: %s" % kws.keys())
