@@ -91,7 +91,7 @@ class TestCase(unittest.TestCase):
 
     _exact_typesets = [(bool, np.bool_), utils.INT_TYPES, (str,), (np.integer,), (utils.text_type), ]
     _approx_typesets = [(float,), (complex,), (np.inexact)]
-    _sequence_typesets = [(tuple,)]
+    _sequence_typesets = [(tuple, list)]
     _float_types = (float, np.floating)
     _complex_types = (complex, np.complexfloating)
 
@@ -142,7 +142,6 @@ class TestCase(unittest.TestCase):
             failure_msg = str(exc)
             # Fall off of the 'except' scope to avoid Python 3 exception
             # chaining.
-            self.fail("when comparing %s and %s: %s" % (first, second, failure_msg))
         else:
             return
         # Decorate the failure message with more information
@@ -171,6 +170,8 @@ class TestCase(unittest.TestCase):
         second_family = self._detect_family(second)
 
         assertion_message = "Type Family mismatch. (%s != %s)" % (first_family, second_family)
+        if msg:
+            assertion_message += ': %s' % (msg,)
         self.assertEqual(first_family, second_family, msg=assertion_message)
 
         # We now know they are in the same comparison family
@@ -257,6 +258,17 @@ def override_config(name, value):
         setattr(config, name, old_value)
 
 
+def compile_function(name, code, globs):
+    """
+    Given a *code* string, compile it with globals *globs* and return
+    the function named *name*.
+    """
+    co = compile(code, "<string>", "exec")
+    ns = {}
+    eval(co, globs, ns)
+    return ns[name]
+
+
 # From CPython
 
 @contextlib.contextmanager
