@@ -4,7 +4,7 @@ import numpy as np
 
 import numba.unittest_support as unittest
 from numba.compiler import compile_isolated, Flags
-from numba import types, utils
+from numba import types, utils, njit
 from numba.tests import usecases
 from .support import TestCase
 
@@ -93,6 +93,9 @@ def integer_indexing_1d_usecase(a, i):
 def integer_indexing_2d_usecase(a, i1, i2):
     return a[i1,i2]
 
+def integer_indexing_2d_usecase2(a, i1, i2):
+    return a[i1][i2]
+
 def ellipse_usecase(a):
     return a[...,0]
 
@@ -490,10 +493,10 @@ class TestIndexing(TestCase):
         with self.assertTypingError():
             self.test_integer_indexing_1d_for_2d(flags=Noflags)
 
-    def test_2d_integer_indexing(self, flags=enable_pyobj_flags):
+    def test_2d_integer_indexing(self, flags=enable_pyobj_flags,
+                                 pyfunc=integer_indexing_2d_usecase):
         # C layout
         a = np.arange(100, dtype='i4').reshape(10, 10)
-        pyfunc = integer_indexing_2d_usecase
         arraytype = types.Array(types.int32, 2, 'C')
         cr = compile_isolated(pyfunc, (arraytype, types.int32, types.int32),
                               flags=flags)
@@ -508,7 +511,6 @@ class TestIndexing(TestCase):
         self.assertFalse(a.flags['C_CONTIGUOUS'])
         self.assertFalse(a.flags['F_CONTIGUOUS'])
 
-        pyfunc = integer_indexing_2d_usecase
         arraytype = types.Array(types.int32, 2, 'A')
         cr = compile_isolated(pyfunc, (arraytype, types.int32, types.int32),
                               flags=flags)
@@ -521,6 +523,23 @@ class TestIndexing(TestCase):
     def test_2d_integer_indexing_npm(self):
         self.test_2d_integer_indexing(flags=Noflags)
 
+    def test_2d_integer_indexing2(self):
+        self.test_2d_integer_indexing(pyfunc=integer_indexing_2d_usecase2)
+        self.test_2d_integer_indexing(flags=Noflags,
+                                      pyfunc=integer_indexing_2d_usecase2)
+
+    def test_2d_integer_indexing_via_call(self):
+        @njit
+        def index1(X, i0):
+            return X[i0]
+        @njit
+        def index2(X, i0, i1):
+            return index1(X[i0], i1)
+        a = np.arange(10).reshape(2, 5)
+        self.assertEqual(index2(a, 0, 0), a[0][0])
+        self.assertEqual(index2(a, 1, 1), a[1][1])
+        self.assertEqual(index2(a, -1, -1), a[-1][-1])
+
     def test_2d_float_indexing(self, flags=enable_pyobj_flags):
         a = np.arange(100, dtype='i4').reshape(10, 10)
         pyfunc = integer_indexing_2d_usecase
