@@ -10,11 +10,19 @@ from numba import unittest_support as unittest
 
 
 @njit
-def f(a, b, c):
-    return a + 2 * b - c
+def f_inner(a, b, c):
+    return a, b, c
+
+def f(x, y, z):
+    return f_inner(x, c=y, b=z)
+
+@njit
+def g_inner(a, b=2, c=3):
+    return a, b, c
 
 def g(x, y, z):
-    return f(x, c=y, b=z)
+    return g_inner(x, b=y), g_inner(a=z, c=x)
+
 
 
 class TestNestedCall(unittest.TestCase):
@@ -38,6 +46,14 @@ class TestNestedCall(unittest.TestCase):
         """
         Test a nested function call with named (keyword) arguments.
         """
+        cfunc = njit(f)
+        self.assertEqual(cfunc(1, 2, 3), f(1, 2, 3))
+        self.assertEqual(cfunc(1, y=2, z=3), f(1, 2, 3))
+
+    def test_default_args(self):
+        """
+        Test a nested function call using default argument values.
+        """
         cfunc = njit(g)
         self.assertEqual(cfunc(1, 2, 3), g(1, 2, 3))
         self.assertEqual(cfunc(1, y=2, z=3), g(1, 2, 3))
