@@ -43,8 +43,6 @@ class TestArgRetCasting(unittest.TestCase):
         typemap = cres.type_annotation.typemap
         # Argument "iters" must be uint32
         self.assertEqual(typemap['iters'], types.uint32)
-        # Localized "iters" must be uint32
-        self.assertEqual(typemap['iters.1'], types.uint32)
 
 
 class TestTupleUnify(unittest.TestCase):
