@@ -3,6 +3,7 @@ from __future__ import print_function, absolute_import
 from collections import defaultdict
 import functools
 import types as pytypes
+import sys
 import weakref
 
 import numpy
@@ -10,10 +11,13 @@ import numpy
 from numba import types
 from numba.typeconv import rules
 from . import templates
+
 # Initialize declarations
-from . import builtins, cmathdecl, mathdecl, npdatetime, npydecl, operatordecl
+from . import (
+    builtins, cmathdecl, mathdecl, npdatetime, npydecl, operatordecl,
+    randomdecl)
 from numba import numpy_support, utils
-from . import ctypes_utils, cffi_utils
+from . import ctypes_utils, cffi_utils, bufproto
 
 
 class BaseContext(object):
@@ -49,6 +53,10 @@ class BaseContext(object):
             raise NotImplementedError(type(num), num)
 
     def resolve_function_type(self, func, args, kws):
+        """
+        Resolve function type *func* for argument types *args* and *kws*.
+        A signature is returned.
+        """
         if isinstance(func, types.Function):
             return func.template(self).apply(args, kws)
 
@@ -56,6 +64,10 @@ class BaseContext(object):
             template, args, kws = func.overloaded.get_call_template(args, kws)
             return template(self).apply(args, kws)
 
+        if isinstance(func, types.ExceptionType):
+            return_type = types.ExceptionInstance(func.exc_class)
+            return templates.signature(return_type)
+
         defns = self.functions[func]
         for defn in defns:
             res = defn.apply(args, kws)
@@ -82,7 +94,8 @@ class BaseContext(object):
                 raise
 
         ret = attrinfo.resolve(value, attr)
-        assert ret
+        if ret is None:
+            raise KeyError(attr)
         return ret
 
     def resolve_setitem(self, target, index, value):
@@ -117,10 +130,6 @@ class BaseContext(object):
         if isinstance(val, utils.INT_TYPES):
             # Force all integers to be 64-bit
             return types.int64
-        elif numpy_support.is_array(val):
-            dtype = numpy_support.from_dtype(val.dtype)
-            layout = numpy_support.map_layout(val)
-            return types.Array(dtype, val.ndim, layout)
 
         tp = self.resolve_data_type(val)
         if tp is None:
@@ -162,27 +171,46 @@ class BaseContext(object):
             else:
                 return types.Tuple(tys)
 
-        else:
-            try:
-                return numpy_support.map_arrayscalar_type(val)
-            except NotImplementedError:
-                pass
+        elif ctypes_utils.is_ctypes_funcptr(val):
+            return ctypes_utils.make_function_type(val)
 
-        if numpy_support.is_array(val):
+        elif cffi_utils.SUPPORTED and cffi_utils.is_cffi_func(val):
+            return cffi_utils.make_function_type(val)
+
+        elif numpy_support.is_array(val):
             ary = val
             try:
                 dtype = numpy_support.from_dtype(ary.dtype)
             except NotImplementedError:
                 return
-
-            if ary.flags.c_contiguous:
-                layout = 'C'
-            elif ary.flags.f_contiguous:
-                layout = 'F'
-            else:
-                layout = 'A'
+            layout = numpy_support.map_layout(ary)
             return types.Array(dtype, ary.ndim, layout)
 
+        elif sys.version_info >= (2, 7) and not isinstance(val, numpy.generic):
+            try:
+                m = memoryview(val)
+            except TypeError:
+                pass
+            else:
+                # Object has the buffer protocol
+                try:
+                    dtype = bufproto.decode_pep3118_format(m.format, m.itemsize)
+                except ValueError:
+                    pass
+                else:
+                    type_class = bufproto.get_type_class(type(val))
+                    layout = bufproto.infer_layout(m)
+                    return type_class(dtype, m.ndim, layout=layout,
+                                      readonly=m.readonly)
+
+        else:
+            # Matching here is quite broad, so we have to do it after
+            # the more specific matches above.
+            try:
+                return numpy_support.map_arrayscalar_type(val)
+            except NotImplementedError:
+                pass
+
         return
 
     def resolve_value_type(self, val):
@@ -194,19 +222,11 @@ class BaseContext(object):
         if tp is not None:
             return tp
 
-        elif ctypes_utils.is_ctypes_funcptr(val):
-            cfnptr = val
-            return ctypes_utils.make_function_type(cfnptr)
-
-        elif cffi_utils.SUPPORTED and cffi_utils.is_cffi_func(val):
-            return cffi_utils.make_function_type(val)
-
-        elif (cffi_utils.SUPPORTED and
-                  isinstance(val, cffi_utils.ExternCFunction)):
+        elif isinstance(val, types.ExternalFunction):
             return val
 
-        elif type(val) is type and issubclass(val, BaseException):
-            return types.exception_type
+        elif isinstance(val, type) and issubclass(val, BaseException):
+            return types.ExceptionType(val)
 
         else:
             try:
@@ -391,6 +411,7 @@ class Context(BaseContext):
         self.install(mathdecl.registry)
         self.install(npydecl.registry)
         self.install(operatordecl.registry)
+        self.install(randomdecl.registry)
 
 
 def new_method(fn, sig):
