@@ -2,7 +2,12 @@
 Define typing templates
 """
 from __future__ import print_function, division, absolute_import
-from numba import types
+
+import functools
+from .. import types
+from ..typeinfer import TypingError
+from functools import reduce
+import operator
 
 
 class Signature(object):
@@ -13,12 +18,25 @@ class Signature(object):
         self.args = args
         self.recvr = recvr
 
+    def __getstate__(self):
+        """
+        Needed because of __slots__.
+        """
+        return self.return_type, self.args, self.recvr
+
+    def __setstate__(self, state):
+        """
+        Needed because of __slots__.
+        """
+        self.return_type, self.args, self.recvr = state
+
     def __hash__(self):
-        return hash(self.args)
+        return hash((self.args, self.return_type))
 
     def __eq__(self, other):
         if isinstance(other, Signature):
             return (self.args == other.args and
+                    self.return_type == other.return_type and
                     self.recvr == other.recvr)
 
     def __ne__(self, other):
@@ -73,37 +91,54 @@ class Rating(object):
         """
         return (self.unsafe_convert, self.safe_convert, self.promote)
 
+    def __add__(self, other):
+        if type(self) is not type(other):
+            return NotImplemented
+        rsum = Rating()
+        rsum.promote = self.promote + other.promote
+        rsum.safe_convert = self.safe_convert + other.safe_convert
+        rsum.unsafe_convert = self.unsafe_convert + other.unsafe_convert
+        return rsum
+
+
+def _rate_arguments(context, actualargs, formalargs):
+    ratings = [Rating()]
+    for actual, formal in zip(actualargs, formalargs):
+        rate = Rating()
+        by = context.type_compatibility(actual, formal)
+        if by is None:
+            return None
+
+        if by == 'promote':
+            rate.promote += 1
+        elif by == 'safe':
+            rate.safe_convert += 1
+        elif by == 'unsafe':
+            rate.unsafe_convert += 1
+        elif by == 'exact':
+            pass
+        else:
+            raise Exception("unreachable", by)
+
+        ratings.append(rate)
+    return ratings
+
 
 def resolve_overload(context, key, cases, args, kws):
     assert not kws, "Keyword arguments are not supported, yet"
     # Rate each cases
     candids = []
-    ratings = []
+    symm_ratings = []
     for case in cases:
         if len(args) == len(case.args):
-            rate = Rating()
-            for actual, formal in zip(args, case.args):
-                by = context.type_compatibility(actual, formal)
-                if by is None:
-                    break
-
-                if by == 'promote':
-                    rate.promote += 1
-                elif by == 'safe':
-                    rate.safe_convert += 1
-                elif by == 'unsafe':
-                    rate.unsafe_convert += 1
-                elif by == 'exact':
-                    pass
-                else:
-                    raise Exception("unreachable", by)
-
-            else:
-                ratings.append(rate.astuple())
+            ratings = _rate_arguments(context, args, case.args)
+            if ratings is not None:
+                combined = reduce(operator.add, ratings)
+                symm_ratings.append(combined.astuple())
                 candids.append(case)
 
     # Find the best case
-    ordered = sorted(zip(ratings, candids), key=lambda i: i[0])
+    ordered = sorted(zip(symm_ratings, candids), key=lambda i: i[0])
     if ordered:
         if len(ordered) > 1:
             (first, case1), (second, case2) = ordered[:2]
@@ -119,7 +154,7 @@ def resolve_overload(context, key, cases, args, kws):
 
                 # Try to resolve promotion
                 # TODO: need to match this to the C overloading dispatcher
-                resolvable = resolve_ambiguous_promotions(context, ambiguous,
+                resolvable = resolve_ambiguous_resolution(context, ambiguous,
                                                           args)
                 if resolvable:
                     return resolvable
@@ -132,47 +167,15 @@ def resolve_overload(context, key, cases, args, kws):
         return ordered[0][1]
 
 
-class UnsafePromotionError(Exception):
-    pass
-
-
-def safe_promotion(actual, formal):
+def resolve_ambiguous_resolution(context, cases, args):
+    """Uses asymmetric resolution to find the best version
     """
-    Allow integer to be casted to the nearest integer
-    """
-    # Integers?
-    if actual in types.integer_domain and formal in types.integer_domain:
-        # Same signedness?
-        if actual.signed == formal.signed:
-            # Score by their distance
-            return formal.bitwidth - actual.bitwidth
-    raise UnsafePromotionError(actual, formal)
-
-
-def resolve_ambiguous_promotions(context, cases, args):
     ratings = []
     for case in cases:
-        try:
-            rate = _safe_promote_case(context, case, args)
-        except UnsafePromotionError:
-            # Ignore error
-            pass
-        else:
-            ratings.append((rate, case))
+        rates = _rate_arguments(context, args, case.args)
+        ratings.append((tuple(r.astuple() for r in rates), case))
 
-    _, bestcase = min(ratings)
-    return bestcase
-
-
-def _safe_promote_case(context, case, args):
-    rate = 0
-    for actual, formal in zip(args, case.args):
-        by = context.type_compatibility(actual, formal)
-        if by == 'promote':
-            rate += safe_promotion(actual, formal)
-        else:
-            raise UnsafePromotionError(actual, formal)
-    return rate
+    return max(ratings, key=lambda x: x[0])[1]
 
 
 class FunctionTemplate(object):
@@ -194,6 +197,19 @@ class AbstractTemplate(FunctionTemplate):
     def apply(self, args, kws):
         generic = getattr(self, "generic")
         sig = generic(args, kws)
+
+        # Unpack optional type if no matching signature
+        if not sig and any(isinstance(x, types.Optional) for x in args):
+            def unpack_opt(x):
+                if isinstance(x, types.Optional):
+                    return x.type
+                else:
+                    return x
+
+            args = list(map(unpack_opt, args))
+            assert not kws  # Not supported yet
+            sig = generic(args, kws)
+
         if sig:
             cases = [sig]
             return self._select(cases, args, kws)
@@ -211,16 +227,29 @@ class ConcreteTemplate(FunctionTemplate):
         return self._select(cases, args, kws)
 
 
+class UntypedAttributeError(TypingError):
+    def __init__(self, value, attr):
+        msg = 'Unknown attribute "{attr}" of type {type}'.format(type=value,
+                                                              attr=attr)
+        super(UntypedAttributeError, self).__init__(msg)
+
+
 class AttributeTemplate(object):
     def __init__(self, context):
         self.context = context
 
     def resolve(self, value, attr):
+        ret = self._resolve(value, attr)
+        if ret is None:
+            raise UntypedAttributeError(value=value, attr=attr)
+        return ret
+
+    def _resolve(self, value, attr):
         fn = getattr(self, "resolve_%s" % attr, None)
         if fn is None:
             fn = self.generic_resolve
             if fn is NotImplemented:
-                raise NotImplementedError(value, attr)
+                return self.context.resolve_module_constants(value, attr)
             else:
                 return fn(value, attr)
         else:
@@ -229,6 +258,39 @@ class AttributeTemplate(object):
     generic_resolve = NotImplemented
 
 
+def bound_function(template_key):
+    """
+    Wrap an AttributeTemplate resolve_* method to allow it to
+    resolve an instance method's signature rather than a instance attribute.
+    The wrapped method must return the resolved method's signature
+    according to the given self type, args, and keywords.
+
+    It is used thusly:
+
+        class ComplexAttributes(AttributeTemplate):
+            @bound_function("complex.conjugate")
+            def resolve_conjugate(self, ty, args, kwds):
+                return ty
+
+    *template_key* (e.g. "complex.conjugate" above) will be used by the
+    target to look up the method's implementation, as a regular function.
+    """
+    def wrapper(method_resolver):
+        @functools.wraps(method_resolver)
+        def attribute_resolver(self, ty):
+            class MethodTemplate(AbstractTemplate):
+                key = template_key
+                def generic(_, args, kws):
+                    sig = method_resolver(self, ty, args, kws)
+                    if sig is not None:
+                        sig.recvr = ty
+                        return sig
+
+            return types.BoundFunction(MethodTemplate, ty)
+        return attribute_resolver
+    return wrapper
+
+
 class ClassAttrTemplate(AttributeTemplate):
     def __init__(self, context, key, clsdict):
         super(ClassAttrTemplate, self).__init__(context)
@@ -264,6 +326,27 @@ class Registry(object):
     def register_global(self, v, t):
         self.globals.append((v, t))
 
+    def resolves_global(self, global_value, wrapper_type=types.Function,
+                        typing_key=None):
+        """
+        Decorate a FunctionTemplate subclass so that it gets registered
+        as resolving *global_value* with the *wrapper_type* (by default
+        a types.Function).
+
+        Example use::
+            @resolves_global(math.fabs)
+            class Math(ConcreteTemplate):
+                cases = [signature(types.float64, types.float64)]
+        """
+        if typing_key is None:
+            typing_key = global_value
+        def decorate(cls):
+            class Template(cls):
+                key = typing_key
+            self.register_global(global_value, wrapper_type(Template))
+            return cls
+        return decorate
+
 
 builtin_registry = Registry()
 builtin = builtin_registry.register
